/* $Id: stubs-a8-z8-lib.h,v 1.9 2002/09/08 03:05:46 skids Exp $
******************************************************************************

   Inlines and typedefs for stubs version of emulated alpha/z support.

   Copyright (C) 2001 Brian S. Julin [bri@tull.umassp.edu]

   Permission is hereby granted, free of charge, to any person obtaining a
   copy of this software and associated documentation files (the "Software"),
   to deal in the Software without restriction, including without limitation
   the rights to use, copy, modify, merge, publish, distribute, sublicense,
   and/or sell copies of the Software, and to permit persons to whom the
   Software is furnished to do so, subject to the following conditions:

   The above copyright notice and this permission notice shall be included in
   all copies or substantial portions of the Software.

   THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
   IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
   FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.  IN NO EVENT SHALL
   THE AUTHOR(S) BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER
   IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
   CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.

******************************************************************************
*/

#include <ggi/internal/buf.h>
#include <ggi/internal/ggi-dl.h>

static inline void do_alpha_over_div(ggi_visual *vis,
				     ggi_pixel srcp, ggi_bufval srca,
				     ggi_pixel *dstp_p, uint8 *dsta_p)
{
	u_int32_t term1, term2, term3;
	ggi_color src, dst;

	ggiUnmapPixel(vis, srcp, &src);
	ggiUnmapPixel(vis, *dstp_p, &dst);

	term1 = (255 - srca) * *dsta_p;

	term2 = srca * 255 + term1;  /* == 65025 * resultant alpha */

	if (term2 == 0) return;	     /* Punt! Both transparent. */

	term3 = 	             /* == 65025 * unadjusted resultant r */
	  (uint32)dst.r * term1 + (uint32)srca * src.r * 255;
	dst.r = term3/term2;         /* == adjusted resultant red */

	term3 = 	             /* == 65025 * unadjusted resultant g */
	  (uint32)dst.g * term1 + (uint32)srca * src.g * 255;
	dst.g = term3/term2;         /* == adjusted resultant red */

	term3 = 	             /* == 65025 * unadjusted resultant b */
	  (uint32)dst.b * term1 + (uint32)srca * src.b * 255;
	dst.b = term3/term2;         /* == adjusted resultant red */

	*dsta_p = term2/255;

	*dstp_p = ggiMapColor(vis, &dst);

}

static inline void do_alpha_over(ggi_visual *vis,
				 ggi_pixel srcp, ggi_bufval srca,
				 ggi_pixel *dstp_p, uint8 *dsta_p)
{
	u_int32_t term1, term2;
	ggi_color src, dst;


	ggiUnmapPixel(vis, srcp, &src);
	ggiUnmapPixel(vis, *dstp_p, &dst);

	term1 = (255 - srca) * *dsta_p;

	term2 = srca * 255 + term1;  /* == 65025 * resultant alpha */
	if (term2 == 0) return;	     /* Punt! Both transparent. */

	dst.r = ((uint32)dst.r * term1 + (uint32)srca * src.r * 255)/65025;
	dst.g = ((uint32)dst.g * term1 + (uint32)srca * src.g * 255)/65025;
	dst.b = ((uint32)dst.b * term1 + (uint32)srca * src.b * 255)/65025;

	*dsta_p = term2/255;
	
	*dstp_p = ggiMapColor(vis, &dst);
}


static inline void do_alpha_under_div(ggi_visual *vis,
				      ggi_pixel srcp, ggi_bufval srca,
				      ggi_pixel *dstp_p, uint8 *dsta_p)
{
	int dsta;
	u_int32_t term1, term2, term3;
	ggi_color src, dst;

	dsta = *dsta_p;

	ggiUnmapPixel(vis, srcp, &src);
	ggiUnmapPixel(vis, *dstp_p, &dst);

	term1 = (255 - dsta) * srca;

	term2 = dsta * 255 + term1;  /* == 65025 * resultant alpha */
	if (term2 == 0) return;	     /* Punt! Both transparent. */

	term3 = 	             /* == 65025 * unadjusted resultant red */
	  (uint32)src.r * term1 + (uint32)dsta * dst.r * 255;
	src.r = term3/term2;         /* == adjusted resultant red */

	term3 = 	             /* == 65025 * unadjusted resultant red */
	  (uint32)src.g * term1 + (uint32)dsta * dst.g * 255;
	src.g = term3/term2;         /* == adjusted resultant red */

	term3 = 	             /* == 65025 * unadjusted resultant red */
	  (uint32)src.b * term1 + (uint32)dsta * dst.b * 255;
	src.b = term3/term2;         /* == adjusted resultant red */

	*dsta_p = term2/255;

	*dstp_p = ggiMapColor(vis, &src);
}


static inline void do_alpha_under(ggi_visual *vis,
				  ggi_pixel srcp, ggi_bufval srca,
				  ggi_pixel *dstp_p, uint8 *dsta_p)
{
	int dsta;
	u_int32_t term1, term2;
	ggi_color src, dst;

	dsta = *dsta_p;

	ggiUnmapPixel(vis, srcp, &src);
	ggiUnmapPixel(vis, *dstp_p, &dst);

	term1 = (255 - dsta) * srca;

	term2 = dsta * 255 + term1;  /* == 65025 * resultant alpha */
	if (term2 == 0) return;	     /* Punt! Both transparent. */

	src.r = ((uint32)src.r * term1 + (uint32)dsta * dst.r * 255)/65025;
	src.g = ((uint32)src.g * term1 + (uint32)dsta * dst.g * 255)/65025;
	src.b = ((uint32)src.b * term1 + (uint32)dsta * dst.b * 255)/65025;

	*dsta_p = term2/255;

	*dstp_p = ggiMapColor(vis, &src);
}

static inline void do_alpha_in_div(ggi_visual *vis,
				   ggi_pixel srcp, ggi_bufval srca,
				   ggi_pixel *dstp_p, uint8 *dsta_p)
{
	uint32 term1;
	ggi_color src;

	ggiUnmapPixel(vis, srcp, &src);

	term1 = *dsta_p * srca;	     /* == 65025 * resultant alpha */
	if (!term1) {
	  src.r = src.g = src.b = 0;
	  *dsta_p = 0;
	} else {
	  *dsta_p = term1/255;
	}

	*dstp_p = ggiMapColor(vis, &src);
}

static inline void do_alpha_in(ggi_visual *vis,
			       ggi_pixel srcp, ggi_bufval srca,
			       ggi_pixel *dstp_p, uint8 *dsta_p)
{
	uint32 term1;
	ggi_color src;

	ggiUnmapPixel(vis, srcp, &src);

	term1 = *dsta_p * srca;	     /* == 65025 * resultant alpha */
	src.r = term1 * src.r / 65025;
	src.g = term1 * src.g / 65025;
	src.b = term1 * src.b / 65025;
	*dsta_p = term1/255;

	*dstp_p = ggiMapColor(vis, &src);
}

static inline void do_alpha_dst_in_div(ggi_visual *vis,
				   ggi_pixel srcp, ggi_bufval srca,
				   ggi_pixel *dstp_p, uint8 *dsta_p)
{
	uint32 term1;
	ggi_color dst;

	ggiUnmapPixel(vis, *dstp_p, &dst);

	term1 = srca * *dsta_p;	     /* == 65025 * resultant alpha */
	if (!term1) {
	  dst.r = dst.g = dst.b = 0;
	  *dsta_p = 0;
	} else {
	  *dsta_p = term1/255;
	}


	*dstp_p = ggiMapColor(vis, &dst);

}

static inline void do_alpha_dst_in(ggi_visual *vis,
				   ggi_pixel srcp, ggi_bufval srca,
				   ggi_pixel *dstp_p, uint8 *dsta_p)
{
	uint32 term1;
	ggi_color dst;

	ggiUnmapPixel(vis, *dstp_p, &dst);

	term1 = srca * *dsta_p;	     /* == 65025 * resultant alpha */
	dst.r = term1 * dst.r / 65025;
	dst.g = term1 * dst.g / 65025;
	dst.b = term1 * dst.b / 65025;
	*dsta_p = term1/255;

	*dstp_p = ggiMapColor(vis, &dst);
}

static inline void do_alpha_out_div(ggi_visual *vis,
				    ggi_pixel srcp, ggi_bufval srca,
				    ggi_pixel *dstp_p, uint8 *dsta_p)
{
	uint32 term1;
	ggi_color src;

	ggiUnmapPixel(vis, srcp, &src);

	term1 = (255 - *dsta_p) * srca;	     /* == 65025 * resultant alpha */
	if (!term1) {
	  src.r = src.g = src.b = 0;
	  *dsta_p = 0;
	} else {
	  *dsta_p = term1/255;
	}

	*dstp_p = ggiMapColor(vis, &src);
}

static inline void do_alpha_out(ggi_visual *vis,
				ggi_pixel srcp, ggi_bufval srca,
				ggi_pixel *dstp_p, uint8 *dsta_p)
{
	uint32 term1;
	ggi_color src;

	ggiUnmapPixel(vis, srcp, &src);

	term1 = (255 - *dsta_p) * srca;	     /* == 65025 * resultant alpha */
	src.r = term1 * src.r / 65025;
	src.g = term1 * src.g / 65025;
	src.b = term1 * src.b / 65025;
	*dsta_p = term1/255;

	*dstp_p = ggiMapColor(vis, &src);
}

static inline void do_alpha_dst_out_div(ggi_visual *vis,
					ggi_pixel srcp, ggi_bufval srca,
					ggi_pixel *dstp_p, uint8 *dsta_p)
{
	uint32 term1;
	ggi_color dst;

	ggiUnmapPixel(vis, *dstp_p, &dst);

	term1 = (255 - srca) * *dsta_p;	     /* == 65025 * resultant alpha */
	if (!term1) {
	  dst.r = dst.g = dst.b = 0;
	  *dsta_p = 0;
	} else {
	  *dsta_p = term1/255;
	}

	*dstp_p = ggiMapColor(vis, &dst);
}

static inline void do_alpha_dst_out(ggi_visual *vis,
				    ggi_pixel srcp, ggi_bufval srca,
				    ggi_pixel *dstp_p, uint8 *dsta_p)
{
	uint32 term1;
	ggi_color dst;

	ggiUnmapPixel(vis, *dstp_p, &dst);

	term1 = (255 - srca) * *dsta_p;	     /* == 65025 * resultant alpha */
	dst.r = term1 * dst.r / 65025;
	dst.g = term1 * dst.g / 65025;
	dst.b = term1 * dst.b / 65025;

	*dsta_p = term1/255;

	*dstp_p = ggiMapColor(vis, &dst);
}

static inline void do_alpha_top_div(ggi_visual *vis,
				    ggi_pixel srcp, ggi_bufval srca,
				    ggi_pixel *dstp_p, uint8 *dsta_p)
{
	int dsta;
	u_int32_t term1, term2, term3;
	ggi_color src, dst;

	ggiUnmapPixel(vis, srcp, &src);
	ggiUnmapPixel(vis, *dstp_p, &dst);

	dsta = *dsta_p;

	term1 = (255 - srca) * dsta;

	term2 = srca * dsta + term1;  /* == 65025 * resultant alpha */
	if (term2 == 0) return;	     /* Punt! Both transparent. */

	term3 = 	             /* == 65025 * unadjusted resultant red */
	  (uint32)dst.r * term1 + (uint32)srca * src.r * dsta;
	dst.r = term3/term2;         /* == adjusted resultant red */

	term3 = 	             /* == 65025 * unadjusted resultant red */
	  (uint32)dst.g * term1 + (uint32)srca * src.g * dsta;
	dst.g = term3/term2;         /* == adjusted resultant red */

	term3 = 	             /* == 65025 * unadjusted resultant red */
	  (uint32)dst.b * term1 + (uint32)srca * src.b * dsta;
	dst.b = term3/term2;         /* == adjusted resultant red */

	*dsta_p = term2/255;

	*dstp_p = ggiMapColor(vis, &dst);
}

static inline void do_alpha_top(ggi_visual *vis,
				ggi_pixel srcp, ggi_bufval srca,
				ggi_pixel *dstp_p, uint8 *dsta_p)
{
	int dsta;
	u_int32_t term1, term2;
	ggi_color src, dst;

	ggiUnmapPixel(vis, srcp, &src);
	ggiUnmapPixel(vis, *dstp_p, &dst);
	
	dsta = *dsta_p;
	
	term1 = (255 - srca) * dsta;
	
	term2 = srca * dsta + term1;  /* == 65025 * resultant alpha */
	if (term2 == 0) return;	     /* Punt! Both transparent. */
	
	dst.r = ((uint32)dst.r * term1 + (uint32)srca * src.r * dsta)/65025;
	dst.g = ((uint32)dst.g * term1 + (uint32)srca * src.g * dsta)/65025;
	dst.b = ((uint32)dst.b * term1 + (uint32)srca * src.b * dsta)/65025;
	
	*dsta_p = term2/255;
	
	*dstp_p = ggiMapColor(vis, &dst);
}


static inline void do_alpha_bottom_div(ggi_visual *vis,
				       ggi_pixel srcp, ggi_bufval srca,
				       ggi_pixel *dstp_p, uint8 *dsta_p)
{
	int dsta;
	u_int32_t term1, term2, term3;
	ggi_color src, dst;

	dsta = *dsta_p;

	ggiUnmapPixel(vis, srcp, &src);
	ggiUnmapPixel(vis, *dstp_p, &dst);

	term1 = (255 - dsta) * srca;

	term2 = dsta * srca + term1;  /* == 65025 * resultant alpha */
	if (term2 == 0) return;	     /* Punt! Both transparent. */

	term3 = 	             /* == 65025 * unadjusted resultant red */
	  (uint32)src.r * term1 + (uint32)dsta * dst.r * srca;
	src.r = term3/term2;         /* == adjusted resultant red */

	term3 = 	             /* == 65025 * unadjusted resultant red */
	  (uint32)src.g * term1 + (uint32)dsta * dst.g * srca;
	src.g = term3/term2;         /* == adjusted resultant red */

	term3 = 	             /* == 65025 * unadjusted resultant red */
	  (uint32)src.b * term1 + (uint32)dsta * dst.b * srca;
	src.b = term3/term2;         /* == adjusted resultant red */

	*dsta_p = term2/255;

	*dstp_p = ggiMapColor(vis, &src);
}


static inline void do_alpha_bottom(ggi_visual *vis,
				   ggi_pixel srcp, ggi_bufval srca,
				   ggi_pixel *dstp_p, uint8 *dsta_p)
{
	int dsta;
	u_int32_t term1, term2;
	ggi_color src, dst;

	dsta = *dsta_p;

	ggiUnmapPixel(vis, srcp, &src);
	ggiUnmapPixel(vis, *dstp_p, &dst);

	term1 = (255 - dsta) * srca;

	term2 = dsta * srca + term1;  /* == 65025 * resultant alpha */
	if (term2 == 0) return;	     /* Punt! Both transparent. */

	src.r = ((uint32)src.r * term1 + (uint32)dsta * dst.r * srca)/65025;
	src.g = ((uint32)src.g * term1 + (uint32)dsta * dst.g * srca)/65025;
	src.b = ((uint32)src.b * term1 + (uint32)dsta * dst.b * srca)/65025;

	*dsta_p = term2/255;

	*dstp_p = ggiMapColor(vis, &src);
}

static inline void do_alpha_xor_div(ggi_visual *vis,
				    ggi_pixel srcp, ggi_bufval srca,
				    ggi_pixel *dstp_p, uint8 *dsta_p)
{
	int _dsta;
	u_int32_t term1, term2, term3;
	ggi_color src, dst;

	ggiUnmapPixel(vis, srcp, &src);
	ggiUnmapPixel(vis, *dstp_p, &dst);

	term1 = (255 - srca) * *dsta_p;
	_dsta = (255 - *dsta_p);

	term2 = srca * _dsta + term1; /* == 65025 * resultant alpha */
	if (term2 == 0) return;	     /* Punt! Both transparent. */

	term3 = 	             /* == 65025 * unadjusted resultant red */
	  (uint32)dst.r * term1 + (uint32)srca * src.r * _dsta;
	dst.r = term3/term2;         /* == adjusted resultant red */

	term3 = 	             /* == 65025 * unadjusted resultant red */
	  (uint32)dst.g * term1 + (uint32)srca * src.g * _dsta;
	dst.g = term3/term2;         /* == adjusted resultant red */

	term3 = 	             /* == 65025 * unadjusted resultant red */
	  (uint32)dst.b * term1 + (uint32)srca * src.b * _dsta;
	dst.b = term3/term2;         /* == adjusted resultant red */

	*dsta_p = term2/255;

	*dstp_p = ggiMapColor(vis, &dst);
}

static inline void do_alpha_xor(ggi_visual *vis,
				ggi_pixel srcp, ggi_bufval srca,
				ggi_pixel *dstp_p, uint8 *dsta_p)
{
	int _dsta;
	u_int32_t term1, term2;
	ggi_color src, dst;

	ggiUnmapPixel(vis, srcp, &src);
	ggiUnmapPixel(vis, *dstp_p, &dst);

	term1 = (255 - srca) * *dsta_p;
	_dsta = (255 - *dsta_p);

	term2 = srca * _dsta + term1; /* == 65025 * resultant alpha */
	if (term2 == 0) return;	     /* Punt! Both transparent. */

	dst.r = ((uint32)dst.r * term1 + (uint32)srca * src.r * _dsta)/65025;
	dst.g = ((uint32)dst.g * term1 + (uint32)srca * src.g * _dsta)/65025;
	dst.b = ((uint32)dst.b * term1 + (uint32)srca * src.b * _dsta)/65025;

	*dsta_p = term2/255;
	
	*dstp_p = ggiMapColor(vis, &dst);
}

static inline void do_alpha_plus_div(ggi_visual *vis,
				     ggi_pixel srcp, ggi_bufval srca,
				     ggi_pixel *dstp_p, uint8 *dsta_p)
{
	int _dsta;
	u_int32_t term1, term2, term3;
	ggi_color src, dst;

	ggiUnmapPixel(vis, srcp, &src);
	ggiUnmapPixel(vis, *dstp_p, &dst);

	term1 = (255 - srca) * *dsta_p;
	_dsta = (255 - *dsta_p);

	term2 = srca * _dsta + term1; /* == 65025 * resultant alpha */
	if (term2 == 0) return;	     /* Punt! Both transparent. */

	term3 = 	             /* == 65025 * unadjusted resultant red */
	  (uint32)dst.r * term1 + (uint32)srca * src.r * _dsta;
	dst.r = term3/term2;         /* == adjusted resultant red */

	term3 = 	             /* == 65025 * unadjusted resultant red */
	  (uint32)dst.g * term1 + (uint32)srca * src.g * _dsta;
	dst.g = term3/term2;         /* == adjusted resultant red */

	term3 = 	             /* == 65025 * unadjusted resultant red */
	  (uint32)dst.b * term1 + (uint32)srca * src.b * _dsta;
	dst.b = term3/term2;         /* == adjusted resultant red */

	*dsta_p = term2/255;

	*dstp_p = ggiMapColor(vis, &dst);
}


static inline void do_alpha_plus(ggi_visual *vis,
				 ggi_pixel srcp, ggi_bufval srca,
				 ggi_pixel *dstp_p, uint8 *dsta_p)
{
	int dsta;
	ggi_color src, dst;

	ggiUnmapPixel(vis, srcp, &src);
	ggiUnmapPixel(vis, *dstp_p, &dst);

	dsta = *dsta_p;
	*dsta_p = dsta * srca / 65025;

	dst.r = ((uint32)dst.r * dsta + (uint32)src.r * srca)/65025;
	dst.g = ((uint32)dst.g * dsta + (uint32)src.g * srca)/65025;
	dst.b = ((uint32)dst.b * dsta + (uint32)src.b * srca)/65025;
	
	*dstp_p = ggiMapColor(vis, &dst);
}

/* For pixel.c and pixela.c */

static inline ggi_pixel calc_pixel(ggi_visual *vis, int x, int y, 
				   ggi_pixel s, ggi_pixel d)
{
	bufpriv *priv;
	uint8 z, defz, sa, da, dstz;
	enum ggiBuf_flags aflags, zflags;
	ggiBuf_t zbuf, abuf;
	ggi_ll_stops_t	 zstops, astops;
	ggi_ll_ststate_t zststate, aststate;

	priv = LIBGGI_BUFEXT(vis);
	abuf = priv->abuf;
	zbuf = priv->zbuf;

	zstops = GGI_LL_STOPS(zbuf);
	zststate = GGI_LL_STSTATE(zbuf);
	zflags = (GGI_LL_MATHGC(zbuf))->math;

	astops = GGI_LL_STOPS(abuf);
	aststate = GGI_LL_STSTATE(abuf);

        zstops->storage2mem(zststate, x, y, 0, &z, 1, 1, 1);
	defz = (GGI_LL_GC(zbuf))->fgsrc;

	dstz = z;
	switch (zflags & BUF_CMP_MASK) {
	case BUF_CMP_EQ:
	  z = (defz == z);
	  break;
	case BUF_CMP_LT:
	  z = (defz < z);
	  break;
	case BUF_CMP_GT:
	  z = (defz > z);
	  break;
	case BUF_CMP_LTE:
	  z = (defz <= z);
	  break;
	case BUF_CMP_GTE:
	  z = (defz >= z);
	  break;
	default:
	  z = 1;
	  break;
	}

	aflags = (GGI_LL_MATHGC(abuf))->math;
	if (!z) { 
	  if (!(aflags & BUF_BLEND_ALT)) {
	    return d;
	  }
	  aflags >>= BUF_ALTBLEND_SHIFT;
	} else {
	  dstz = defz;
	}
	if (!(aflags & BUF_BLEND_POSTDIV)) goto switch2;

        astops->storage2mem(aststate, x, y, 0, &da, 1, 1, 1);
	sa = (GGI_LL_GC(abuf))->fgsrc;

	switch(aflags & BUF_BLENDTYPE_MASK) {
	case BUF_BLEND_OVER:
	  do_alpha_over_div(vis, s, sa, &d, &da);
	  break;
	case BUF_BLEND_UNDER:
	  do_alpha_under_div(vis, s, sa, &d, &da);
	  break;
	case BUF_BLEND_SRC_IN:
	  do_alpha_in_div(vis, s, sa, &d, &da);
	  break;
	case BUF_BLEND_DST_IN:
	  do_alpha_dst_in_div(vis, s, sa, &d, &da);
	  break;
	case BUF_BLEND_SRC_OUT:
	  do_alpha_out_div(vis, s, sa, &d, &da);
	  break;
	case BUF_BLEND_DST_OUT:
	  do_alpha_dst_out_div(vis, s, sa, &d, &da);
	  break;
	case BUF_BLEND_TOP:
	  do_alpha_top_div(vis, s, sa, &d, &da);
	  break;
	case BUF_BLEND_BOTTOM:
	  do_alpha_bottom_div(vis, s, sa, &d, &da);
	  break;
	case BUF_BLEND_XOR:
	  do_alpha_xor_div(vis, s, sa, &d, &da);
	  break;
	case BUF_BLEND_PLUS:
	  do_alpha_plus_div(vis, s, sa, &d, &da);
	  break;
	default:
	  d = s;
	  da = sa;
	  break;
	}

	goto finish;

 switch2:

        astops->storage2mem(aststate, x, y, 0, &da, 1, 1, 1);
	sa = (GGI_LL_GC(abuf))->fgsrc;

	switch(aflags & BUF_BLENDTYPE_MASK) {
	case BUF_BLEND_OVER:
	  do_alpha_over(vis, s, sa, &d, &da);
	  break;
	case BUF_BLEND_UNDER:
	  do_alpha_under(vis, s, sa, &d, &da);
	  break;
	case BUF_BLEND_SRC_IN:
	  do_alpha_in(vis, s, sa, &d, &da);
	  break;
	case BUF_BLEND_DST_IN:
	  do_alpha_dst_in(vis, s, sa, &d, &da);
	  break;
	case BUF_BLEND_SRC_OUT:
	  do_alpha_out(vis, s, sa, &d, &da);
	  break;
	case BUF_BLEND_DST_OUT:
	  do_alpha_dst_out(vis, s, sa, &d, &da);
	  break;
	case BUF_BLEND_TOP:
	  do_alpha_top(vis, s, sa, &d, &da);
	  break;
	case BUF_BLEND_BOTTOM:
	  do_alpha_bottom(vis, s, sa, &d, &da);
	  break;
	case BUF_BLEND_XOR:
	  do_alpha_xor(vis, s, sa, &d, &da);
	  break;
	case BUF_BLEND_PLUS:
	  do_alpha_plus(vis, s, sa, &d, &da);
	  break;
	default:
	  d = s;
	  da = sa;
	  break;
	}

 finish:
	if (!(aflags & BUF_DISPOSE)) {
	  astops->mem2storage(aststate, x, y, 0, &da, 1, 1, 1);
	}
	if (!(zflags & BUF_DISPOSE)) {
	  zstops->mem2storage(zststate, x, y, 0, &dstz, 1, 1, 1);
	}
	return d;
}


ggifunc_drawpixel_nc    GGI_stubs_a8_z8_drawpixel_nc;
ggifunc_drawpixel       GGI_stubs_a8_z8_drawpixel;
ggifunc_putpixel_nc     GGI_stubs_a8_z8_putpixel_nc;
ggifunc_putpixel        GGI_stubs_a8_z8_putpixel;

ggifunc_drawpixel_nc    GGI_stubs_a8_z8_drawpixel_nca;
ggifunc_drawpixel       GGI_stubs_a8_z8_drawpixela;
ggifunc_putpixel_nc     GGI_stubs_a8_z8_putpixel_nca;
ggifunc_putpixel        GGI_stubs_a8_z8_putpixela;

ggifunc_copybox         GGI_stubs_a8_z8_copybox;
ggifunc_crossblit       GGI_stubs_a8_z8_crossblit;
